#ifndef _LIBHX_H
#define _LIBHX_H 20110115

/*
 * This file is for programs that really need about everything. The presence of
 * this file is undocumented, but you can rely on its presence and that it
 * always includes a given feature set. (Once an "#include" is added, it shall
 * be considered documented behavior that feature xyz is available merely by
 * inclusion of this file. Henceforth, "#includes" may not be removed from here
 * unless the implementation really went away.
 */

#include <libHX/ctype_helper.h>
#include <libHX/defs.h>
#include <libHX/deque.h>
#include <libHX/list.h>
#include <libHX/init.h>
#include <libHX/io.h>
#include <libHX/map.h>
#include <libHX/misc.h>
#include <libHX/option.h>
#include <libHX/proc.h>
#include <libHX/string.h>

/* not libxml_helper, because that requires libxml2 */

#endif /* _LIBHX_H */
